package de.fhhannover.inform.trust.ifmapj.channel;

import java.io.ByteArrayInputStream;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;

import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;

import de.fhhannover.inform.trust.ifmapj.exception.CommunicationException;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;
import de.fhhannover.inform.trust.ifmapj.log.IfmapJLog;

/**
 * Attempt to create a {@link CommunicationHandler} implementation to be
 * independent of any extra library... Meaning making usage only of standard
 * JAVA classes.
 * 
 * @author aw
 */
class JavaCommunicationHandler extends AbstractCommunicationHandler {
	
	private SSLSocket mSocket;
	private OutputStream mOutputStream;
	private InputStream mInputStream;
	private DataOutput mDataOutputStream;
	private DataInput mDataInputStream;
	private Map<String, String> mReceivedHeaders;
	
	private byte[] mBuffer;
	
	private class StatusLine {
		public int statusCode;
		public String reason;
	}
	
	JavaCommunicationHandler(String url, String user, String pass,
			SSLSocketFactory sslSocketFactory) throws InitializationException {
		super(url, user, pass, sslSocketFactory);
		
		// some random initial buffer
		mBuffer = new byte[1024];
		mReceivedHeaders = new HashMap<String, String>(10);
	}
	
	@Override
	public InputStream doActualRequest(InputStream is) throws IOException, CommunicationException {
		StatusLine statusLine = null;
		int length = is.available();
			
		sendHttpBody(is);
		statusLine = readStatusLine();
		checkStatusLine(statusLine);
		mReceivedHeaders = readHeaders();
		length = getContentLength(mReceivedHeaders);
		return readHttpBody(length);
	}
	
	@Override
	protected void prepareCommunication() throws IOException {
		if (mSocket != null)
			return;
		
		mSocket = (SSLSocket)getSocketFactory().createSocket(getUrl().getHost(), getPort());
		mSocket.setTcpNoDelay(true);
		mSocket.setNeedClientAuth(true);
		mOutputStream = mSocket.getOutputStream();
		mInputStream = mSocket.getInputStream();
		mDataOutputStream =  new DataOutputStream(mOutputStream);
		mDataInputStream = new DataInputStream(mInputStream);
	}
	
	@Override
	protected void createPostRequest(String path) throws IOException {
		mReceivedHeaders.clear();
		writeLine("POST " + path + " HTTP/1.1");
	}

	@Override
	protected void addHeader(String key, String value) throws IOException {
		writeHeaderLine(key, value);
	}

	@Override
	protected void finishHeaders() throws IOException {
		writeHeaderEnding();
	}

	@Override
	protected boolean replyIsGzipped() throws IOException {
		return isGzipCompressed(mReceivedHeaders);
	}

	private void sendHttpBody(InputStream is) throws IOException {
		int length = is.available();
		allocateBuffer(length);
		readStreamIntoBuffer(is, length);
		sendBufferContents(length);
	}
	
	private void writeHeaderLine(String key, String value) throws IOException {
		writeLine(key + ": " + value);
	}

	private void writeHeaderEnding() throws IOException {
		writeLine("");
		mOutputStream.flush();
	}
	
	private void writeLine(String line) throws IOException {
		mDataOutputStream.write((line + "\r\n").getBytes());
	}


	/*
	 * Try closing it under all circumstances. Throw an Exception of the first
	 * error encountered.
	 * 
	 * It's soooo ugly.
	 * 
	 * (non-Javadoc)
	 * @see de.fhhannover.inform.trust.ifmapj.channel.CommunicationHandler#closeTcpConnection()
	 */
	@Override
	public void closeTcpConnection() throws CommunicationException {
		IOException tmpException = null;

		try {
			if (mInputStream != null) mInputStream.close();
		} catch (IOException e) { 
			if (tmpException == null) tmpException = e;
		} finally {
			mInputStream = null;
		}
		
		try {
			if (mOutputStream != null) mOutputStream.close();
		} catch (IOException e) { 
			if (tmpException != null) tmpException = e;
		} finally {
			mOutputStream = null;
		}
		
		try {
			if (mSocket != null) mSocket.close();
		} catch (IOException e) { 
			if (tmpException == null) tmpException = e;
		} finally {
			mSocket = null;
		}
		
		mDataOutputStream = null;
		mDataInputStream = null;
		
		if (tmpException != null)
			throw new CommunicationException(tmpException);
	}


	private boolean isGzipCompressed(Map<String, String> headers) {
		String encoding = null;
		encoding = headers.get("Content-Encoding");
		if (encoding == null)
			encoding = headers.get("Content-encoding");

		if (encoding == null)
			encoding = headers.get("content-encoding");
		
		if (encoding != null)
			return encoding.contains("gzip");
		
		return false;
	}

	private int getContentLength(Map<String, String> headers) 
			throws CommunicationException {
		int ret;
		String lengthHdr = null;
		
		lengthHdr = headers.get("Content-Length");
		
		if (lengthHdr == null)
			lengthHdr = headers.get("Content-length");

		if (lengthHdr == null)
			lengthHdr = headers.get("content-length");
			
		if (lengthHdr == null)
			throw new CommunicationException("No Content-Length header found");
	
		
		try {
			ret = Integer.parseInt(lengthHdr);
		} catch (NumberFormatException e) {
			throw new CommunicationException("Content-Length invalid: "
					+ "\"" + lengthHdr + "\"");
		}
		return ret;
	}

	private Map<String, String> readHeaders() throws CommunicationException, IOException  {
		
		String line = mDataInputStream.readLine();
		HashMap<String, String> ret = new HashMap<String, String>();
		
		while (line != null && line.length() > 0) {
			String fields[] = line.split(":", 2);
			if (fields.length < 2)
				throw new CommunicationException("Invalid Header Received: "  + line);

			// trim a bit
			fields[0] = fields[0].replaceAll("^\\s+", "");
			fields[0] = fields[0].replaceAll("\\s+$", "");
			fields[1] = fields[1].replaceAll("^\\s+", "");
			fields[1] = fields[1].replaceAll("\\s+$", "");
			
			ret.put(fields[0], fields[1]);
			
			line = mDataInputStream.readLine();
		}
		
		if (line == null)
			throw new CommunicationException("Unexpected EOF reached");
		
		return ret;
	}

	private StatusLine readStatusLine() throws CommunicationException, IOException {
		StatusLine ret = new StatusLine();
		String line = mDataInputStream.readLine();
		if (line == null)
			throw new CommunicationException("No status line received");
		
		String fields[] = line.split(" ", 3);
		if (fields.length < 2)
			throw new CommunicationException("Bad status line received");
		String proto = fields[0];
		
		if (!proto.equals("HTTP/1.1"))
			throw new  CommunicationException("Communication not HTTP/1.1");

		try {
			ret.statusCode = Integer.parseInt(fields[1]);
		} catch (NumberFormatException e) {
			throw new CommunicationException("Bad status code received");
		}
	
		if (fields.length == 3)
			ret.reason = fields[2];
		
		return ret;
	}

	private InputStream readHttpBody(int length) throws IOException {
		allocateBuffer(length);
		readStreamIntoBuffer(mInputStream, length);
		return new ByteArrayInputStream(mBuffer, 0, length);
	}

	private void checkStatusLine(StatusLine status) throws CommunicationException {
		if (status.statusCode != 200) {
			IfmapJLog.warn("HTTP Status Code: " 
					+ status.statusCode + " "
					+ status.reason);
			throw new CommunicationException("HTTP Status Code: "
					+ status.statusCode + " " 
					+ status.reason);
		}
	}

	private void sendBufferContents(int length) throws IOException {
		mOutputStream.write(mBuffer, 0, length);
		mOutputStream.flush();
	}

	/**
	 * Helper to read length bytes into {@link JavaCommunicationHandler#mBuffer}.
	 * 
	 * @param is
	 * @param length
	 * @throws IOException
	 */
	private void readStreamIntoBuffer(InputStream is, int length) throws IOException {
		int read = 0;
		int ret;
		
		while (read < length) {
			ret = is.read(mBuffer, read, length - read);
			if (ret == -1)
				throw new IOException("Stream exception");
			read += ret;
		}
	}

	/**
	 * Check if our current buffer is too small, and if yes, allocate some
	 * more
	 * 
	 * @param length
	 */
	private void allocateBuffer(int length) {
		if (length > mBuffer.length)
			mBuffer = new byte[length];
	}

}
