package de.fhhannover.inform.trust.ifmapj.metadata;

import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

import de.fhhannover.inform.trust.ifmapj.binding.IfmapStrings;
import de.fhhannover.inform.trust.ifmapj.log.IfmapJLog;

/**
 * Simple implementation of the {@link StandardIfmapMetadataFactory} interface.
 * 
 * @author aw
 * @author ib
 *
 */
public class StandardIfmapMetadataFactoryImpl implements StandardIfmapMetadataFactory {

	private DocumentBuilder mDocumentBuilder;
	
	public StandardIfmapMetadataFactoryImpl() {
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		dbf.setNamespaceAware(true);
		try {
			mDocumentBuilder = dbf.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			IfmapJLog.error("Could not get DocumentBuilder instance [" +
					e.getMessage() + "]");
			throw new RuntimeException(e);
		}
	}

	@Override
	public Document createIpMac(String startTime, String endTime,
			String dhcpServer) {
		Document doc = createStdSingleElementDocument("ip-mac",
				Cardinality.multiValue);
		
		Element root = (Element)doc.getFirstChild();
		
		appendTextElementIfNotNull(doc, root, "start-time", startTime);
		appendTextElementIfNotNull(doc, root, "end-time", endTime);
		appendTextElementIfNotNull(doc, root, "dhcp-server", dhcpServer);
		
		return doc;
	}

	@Override
	public Document createIpMac() {
		return createIpMac(null, null, null);
	}

	@Override
	public Document createArMac() {
		return createStdSingleElementDocument("access-request-mac",
				Cardinality.singleValue);
	}

	@Override
	public Document createArDev() {
		return createStdSingleElementDocument("access-request-device",
				Cardinality.singleValue);
	}
	
	@Override
	public Document createArIp() {
		return createStdSingleElementDocument("access-request-ip",
				Cardinality.singleValue);
	}

	@Override
	public Document createAuthAs() {
		return createStdSingleElementDocument("authenticated-as",
				Cardinality.singleValue);
	}

	@Override
	public Document createAuthBy() {
		return createStdSingleElementDocument("authenticated-by",
				Cardinality.singleValue);
	}

	@Override
	public Document createDevIp() {
		return createStdSingleElementDocument("device-ip",
				Cardinality.singleValue);
	}

	@Override
	public Document createDiscoveredBy() {
		return createStdSingleElementDocument("discovered-by",
				Cardinality.singleValue);
	}

	@Override
	public Document createRole(String name, String administrativeDomain) {
		Document doc = createStdSingleElementDocument("role",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();
		
		appendTextElementCheckNull(doc, root, "name", name);
		appendTextElementIfNotNull(doc, root, "administrative-domain",
				administrativeDomain);
		
		return doc;
	}
	
	@Override
	public Document createRole(String name) {
		return createRole(name, null);
	}
	
	@Override
	public Document createDevAttr(String name) {
		Document doc = createStdSingleElementDocument("device-attribute",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();
		appendTextElementCheckNull(doc, root, "name", name);
		return doc;
	}
	
	@Override
	public Document createCapability(String name, String administrativeDomain) {
		Document doc = createStdSingleElementDocument("capability",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();
		
		appendTextElementCheckNull(doc, root, "name", name);
		appendTextElementIfNotNull(doc, root, "administrative-domain",
				administrativeDomain);
		
		return doc;
	}
	
	@Override
	public Document createCapability(String name) {
		return createCapability(name, null);
	}
	
	@Override
	public Document createDevChar(String manufacturer, String model, String os,
			String osVersion, String deviceType, String discoveredTime,
			String discovererId, String discoveryMethod) {
		Document doc = createStdSingleElementDocument("device-characteristic",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();
		
		appendTextElementIfNotNull(doc, root, "manufacturer", manufacturer);
		appendTextElementIfNotNull(doc, root, "model", model);
		appendTextElementIfNotNull(doc, root, "os", os);
		appendTextElementIfNotNull(doc, root, "os-version", osVersion);
		appendTextElementIfNotNull(doc, root, "device-type", deviceType);
		
		appendTextElementCheckNull(doc, root, "discovered-time", discoveredTime);
		appendTextElementCheckNull(doc, root, "discovere-id", discovererId);
		appendTextElementCheckNull(doc, root, "discovery-method", discoveryMethod);
		
		return doc;
	}
	
	@Override
	public Document createEnforcementReport(
			EnforcementAction enforcementAction, String otherTypeDefinition,
			String enforcementReason) {
		Document doc = createStdSingleElementDocument("enforcement-report",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();
		
		appendTextElementCheckNull(doc, root, "enforcement-action", enforcementAction);
		
		// TODO Do we need to check consistency with the action?
		appendTextElementIfNotNull(doc, root, "other-type-definition", otherTypeDefinition);
		appendTextElementIfNotNull(doc, root, "enforcement-reason", enforcementReason);
		
		return doc;
	}
	
	@Override
	public Document createEvent(String name, String discoveredTime,
			String discovererId, Integer magnitude, Integer confidence,
			Significance significance, EventType type,
			String otherTypeDefinition, String information,
			String vulnerabilityUri) {
		Document doc = createStdSingleElementDocument("event",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();
		

		appendTextElementCheckNull(doc, root, "name", name);
		appendTextElementCheckNull(doc, root, "discovered-time", discoveredTime);
		appendTextElementCheckNull(doc, root, "disoverer-id", discovererId);
		// TODO add checks for magnitude and confidence for correct range?
		appendTextElementCheckNull(doc, root, "magnitude", magnitude);
		appendTextElementCheckNull(doc, root, "confidence", confidence);
		appendTextElementCheckNull(doc, root, "significance", significance);
		
		appendTextElementIfNotNull(doc, root, "type", type);
		// TODO Do we need to check consistency with the type?
		appendTextElementIfNotNull(doc, root, "other-type-definition", otherTypeDefinition);
		appendTextElementIfNotNull(doc, root, "information", information);
		// TODO Do we need to check consistency with the type?
		appendTextElementIfNotNull(doc, root, "vulnerability-uri", vulnerabilityUri);
		
		return doc;
	}
	
	@Override
	public Document createLayer2Information(Integer vlanNum, String vlanName,
			Integer port, String administrativeDomain) {
		Document doc = createStdSingleElementDocument("layer2-information",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();

		/* non of them is mandatory */
		appendTextElementIfNotNull(doc, root, "vlan", vlanNum);
		appendTextElementIfNotNull(doc, root, "vlan-name", vlanName);
		appendTextElementIfNotNull(doc, root, "port", port);
		appendTextElementIfNotNull(doc, root, "administrative-domain",
				administrativeDomain);
		
		return doc;
	}
	
	@Override
	public Document createLocation(List<LocationInformation> locationInformation, 
			String discoveredTime, String discovererId) {
		Element childEl;
		Document doc = createStdSingleElementDocument("location",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();
		
		if (locationInformation == null || locationInformation.size() == 0)
			throw new NullPointerException("location-information needs to be set for location");
		
		for (LocationInformation tmpLoc : locationInformation) {
			childEl = doc.createElement("location-information");
			childEl.setAttribute("type", tmpLoc.mType);
			childEl.setAttribute("value", tmpLoc.mValue);
			root.appendChild(childEl);
		}
	
		appendTextElementCheckNull(doc, root, "discovered-time", discoveredTime);
		appendTextElementCheckNull(doc, root, "discoverer-id", discovererId);
		return doc;
	}
	
	@Override
	public Document createRequestForInvestigation(String qualifier) {
		Document doc = createStdSingleElementDocument("request-for-investigation",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();
		
		if (qualifier != null)
			root.setAttribute("qualifier", qualifier);
		return doc;
	}
	
	@Override
	public Document createWlanInformation(String ssid,
			List<WlanSecurityType> ssidUnicastSecurity,
			WlanSecurityType ssidGroupSecurity,
			List<WlanSecurityType> ssidManagementSecurity) {
		Element childEl;
		Node txtEl;
		Document doc = createStdSingleElementDocument("wlan-information",
				Cardinality.singleValue);
		Element root = (Element)doc.getFirstChild();
		
		if (ssidUnicastSecurity == null || ssidUnicastSecurity.size() == 0)
			throw new NullPointerException("ssid-unicast-security to be set for wlan-information");

		if (ssidGroupSecurity == null)
			throw new NullPointerException("ssid-group-security to be set for wlan-information");
		
		if (ssidManagementSecurity == null || ssidManagementSecurity.size() == 0)
			throw new NullPointerException("ssid-management-security to be set for wlan-information");
	
		appendTextElementIfNotNull(doc, root, "ssid", ssid);
		
		for (WlanSecurityType wlanSec : ssidUnicastSecurity) {
			childEl = doc.createElement("ssid-unicast-security");
			if (wlanSec.mOtherTypeDefinition != null) {
				childEl.setAttribute("other-type-definition", wlanSec.mOtherTypeDefinition);				
			}
			txtEl = doc.createTextNode(wlanSec.mWlanSecurityEnum.toString());
			childEl.appendChild(txtEl);
			root.appendChild(childEl);
		}
		
		childEl = doc.createElement("ssid-group-security");
		if (ssidGroupSecurity.mOtherTypeDefinition != null) {
			childEl.setAttribute("other-type-definition", ssidGroupSecurity.mOtherTypeDefinition);				
		}
		txtEl = doc.createTextNode(ssidGroupSecurity.mWlanSecurityEnum.toString());
		childEl.appendChild(txtEl);
		root.appendChild(childEl);

		for (WlanSecurityType wlanSec : ssidManagementSecurity) {
			childEl = doc.createElement("ssid-management-security");
			if (wlanSec.mOtherTypeDefinition != null) {
				childEl.setAttribute("other-type-definition", wlanSec.mOtherTypeDefinition);				
			}
			txtEl = doc.createTextNode(wlanSec.mWlanSecurityEnum.toString());
			childEl.appendChild(txtEl);
			root.appendChild(childEl);
		}
		
		return doc;
	}
	
	@Override
	public Document createUnexpectedBehavior(String discoveredTime, String discovererId,
			Integer magnitude, Integer confidence, Significance significance,
			String type) {
		Document doc = createStdSingleElementDocument("unexpected-behavior",
				Cardinality.multiValue);
		Element root = (Element)doc.getFirstChild();
		
		appendTextElementCheckNull(doc, root, "discovered-time", discoveredTime);
		appendTextElementCheckNull(doc, root, "disoverer-id", discovererId);
		// TODO add checks for magnitude and confidence for correct range?
		appendTextElementCheckNull(doc, root, "magnitude", magnitude);
		appendTextElementCheckNull(doc, root, "confidence", confidence);
		appendTextElementCheckNull(doc, root, "significance", significance);
		
		appendTextElementIfNotNull(doc, root, "type", type);
		
		return doc;
	}
	
	private Document createStdSingleElementDocument(String name, Cardinality card) {
		return createSingleElementDocument(
				IfmapStrings.STD_METADATA_PREFIX + ":" + name,
				IfmapStrings.STD_METADATA_NS_URI,
				card);
	}
	
	private Document createSingleElementDocument(String qualifiedName,
			String uri, Cardinality cardinality) {
		Document doc = mDocumentBuilder.newDocument();
		Element e = doc.createElementNS(uri, qualifiedName);
		e.setAttribute("ifmap-cardinality", cardinality.toString());
		doc.appendChild(e);
		return doc;
	}
	
	/**
	 * Helper to create a new element with name elName and append it to the
	 * {@link Element} given by parent if the given value is non-null.
	 * The new {@link Element} will have {@link Text} node containing value.
	 * 
	 * @param doc {@link Document} where parent is located in
	 * @param parent where to append the new element
	 * @param elName the name of the new element.
	 * @param value the value of the {@link Text} node appended to the new element,
	 *        using toString() on the object.
	 */
	private void appendTextElementIfNotNull(Document doc, Element parent,
			String elName, Object value) {
		
			if (value == null)
				return;
			
			appendTextElementCheckNull(doc, parent, elName, value);
			
	}
	
	/**
	 * Helper to create a new element with name elName and append it to the
	 * {@link Element} given by parent. The new {@link Element} will have
	 * {@link Text} node containing value.
	 * 
	 * @param doc {@link Document} where parent is located in
	 * @param parent where to append the new element
	 * @param elName the name of the new element.
	 * @param value the value of the {@link Text} node appended to the new element,
	 *        using toString() on the object.
	 * @throws NullPointerException if value is null, or any other parameter
	 *         is null
	 */
	private void appendTextElementCheckNull(Document doc, Element parent,
			String elName, Object value) {
		
		if (doc == null || parent == null || elName == null)
				throw new NullPointerException("bad parameters given");
		
		if (value == null)
			throw new NullPointerException("null is not allowed for " + elName 
					+ " in " + doc.getFirstChild().getLocalName());
		
		String valueStr = value.toString();
		if (valueStr == null)
			throw new NullPointerException("null-string not allowed for " + elName 
					+ " in " + doc.getFirstChild().getLocalName());
			
		Element child = doc.createElement(elName);
		Text txtCElement = doc.createTextNode(valueStr);
		child.appendChild(txtCElement);
		parent.appendChild(child);
	}
}
